from coverdl.providers.base import Provider
from coverdl.providers.itunes import ITunesProvider
from coverdl.providers.source import Source
from coverdl.utils import get_extension_from_url
from coverdl.cover import ExtCover
import requests

class AppleMusicProvider(Provider):
    SOURCE = Source.APPLE_MUSIC

    def __init__(self):
        super().__init__()
        self.itunes_provider = ITunesProvider()

    def _transform_url(self, itunes_url):
        # Convert the iTunes url into a higher-quality Apple Music equivalent
        cover_url = itunes_url.replace("is1-ssl", "a1")
        cover_url = cover_url.replace("/image/", "/us/")
        cover_url = cover_url.replace("/thumb/", "/r1000/063/")

        cover_url = '/'.join(cover_url.split('/')[:-1])

        return cover_url
    
    def _test_url(self, url):
        r = requests.head(url, timeout=10)

        return r.ok

    def get_covers(self, artist, album) -> list[ExtCover]:
        results = []
        covers = self.itunes_provider.get_covers(artist, album)

        for cover in covers:
            cover.cover_url = self._transform_url(cover.cover_url)
            cover.ext = get_extension_from_url(cover.cover_url)
            cover.source = self.SOURCE
            if self._test_url(cover.cover_url):
                results.append(cover)

        return results



from coverdl.exceptions import ProviderRequestFailed
from coverdl.providers.base import Provider
from coverdl.providers.source import Source
from coverdl.cover import ExtCover
from coverdl.utils import get_extension_from_url, compare
import requests

class ITunesProvider(Provider):
    BASE_URL = "https://itunes.apple.com"
    SOURCE = Source.ITUNES

    def get_covers(self, artist, album) -> list[ExtCover]:
        results = []
        params = {
            "term": artist,
            "media": "music",
            "entity": "album"
        }
        r = requests.get(self.BASE_URL + "/search", params=params, timeout=10)

        if r.ok:
            data = r.json()

            for item in data["results"]:
                # Check if the album names are somewhat similar.
                # TODO: Filter out possible identifiers in the name
                similarity_ratio = compare(item["collectionName"], album)
                if similarity_ratio > 0.8:
                    cover_url = item.get("artworkUrl100") or item.get("artworkUrl60")
                    results.append(
                        ExtCover(
                            artist=item["artistName"],
                            title=item["collectionName"],
                            source=self.SOURCE,
                            cover_url=cover_url,
                            confidence=similarity_ratio,
                            ext=get_extension_from_url(cover_url)
                        )
                    )

            return sorted(results, key=lambda c: c.confidence, reverse=True)
        raise ProviderRequestFailed(self.SOURCE, r.text)



import os
import mimetypes
from coverdl.metadata import SUPPORTED_SONG_EXTENSIONS
from coverdl.cover import Cover
from difflib import SequenceMatcher

IMAGE_EXTENSIONS = [".jpg", ".jpeg", ".png"]

def compare(title1: str, title2: str) -> float:
    return SequenceMatcher(None, title1.lower().strip(), title2.lower().strip()).ratio()

def get_album_paths(path: str, must_have_cover=True) -> list[str]:
    paths = []
    for root, dirs, _ in os.walk(path):
        for current_dir in dirs:
            full_dir = os.path.join(root, current_dir)
            if is_album_dir(full_dir) or (has_song(full_dir) and not must_have_cover):
                paths.append(full_dir)
    return paths

def get_base_path(path: str):
    if os.path.isfile(path):
        return os.path.dirname(path)
    return os.path.normpath(path)

def get_extension_from_url(url: str):
    guessed_type = mimetypes.guess_type(url)

    if guessed_type and len(guessed_type) > 0 and guessed_type[0]:
        return mimetypes.guess_extension(guessed_type[0])

def get_cover(path: str) -> Cover | None:
    if os.path.isfile(path):
        path = os.path.dirname(os.path.abspath(path))

    files = os.listdir(path)

    for file in files:
        name, ext = os.path.splitext(file)
        if ext in IMAGE_EXTENSIONS and name in ["folder", "poster", "cover", "default"]:
            return Cover(path=os.path.join(path, file))

def has_song(dir_path: str):
    for f in os.listdir(dir_path):
        _, ext = os.path.splitext(f)
        if ext in SUPPORTED_SONG_EXTENSIONS:
            return True
    return False

def has_cover(path: str) -> bool:
    return bool(get_cover(path))

def is_album_dir(dir_path: str):
    return has_cover(dir_path) and has_song(dir_path)



from typing import Optional
from abc import ABC, abstractmethod
from coverdl.providers.source import Source
from coverdl.cover import ExtCover

class Provider(ABC):
    BASE_URL: Optional[str] = None
    SOURCE: Source
    API_KEY: Optional[str] = None
    SECRET_KEY: Optional[str] = None

    @abstractmethod
    def get_covers(self, artist: str, album: str) -> list[ExtCover]:
        pass



from coverdl import __version__
from coverdl.cover import ExtCover
from coverdl.exceptions import ProviderRequestFailed
from coverdl.providers.base import Provider
from coverdl.providers.source import Source
from coverdl.utils import get_extension_from_url, compare
import requests

class DiscogsProvider(Provider):
    BASE_URL = "https://api.discogs.com"
    SOURCE = Source.DISCOGS
    API_KEY = "yoVukqDuMTyrckrqJdfc"
    SECRET_KEY = "PRCvdLuRMVghFrNtRRvylkDZEKCiLUbI" # not secret

    def get_headers(self):
        return {
            "User-Agent": f"coverdl/{__version__}",
            "Authorization": f"Discogs key={self.API_KEY}, secret={self.SECRET_KEY}"
        }

    def get_covers(self, artist, album) -> list[ExtCover]:
        params = {
            "artist": artist,
            "release_title": album
        }
        r = requests.get(self.BASE_URL + "/database/search",
                         headers=self.get_headers(),
                         params=params,
                         timeout=10)

        if r.ok:
            results = []
            data = r.json()

            for item in data["results"]:
                similarity_ratio = compare(item["title"], f"{artist} - {album}")

                if similarity_ratio > 0.8:
                    results.append(
                        ExtCover(
                            artist=artist,
                            title=album,
                            source=self.SOURCE,
                            cover_url=item["cover_image"],
                            ext=get_extension_from_url(item["cover_image"]),
                            confidence=similarity_ratio
                        )
                    )

            return sorted(results, key=lambda c: c.confidence, reverse=True)
        raise ProviderRequestFailed(self.SOURCE, r.text)



from PIL import Image
from coverdl.providers.source import Source
import imagehash
import requests
import os
import io

DEFAULT_HEADERS = {
    "User-Agent": "Mozilla/5.0 (Linux; Android 7.0; BLN-L22) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.89 Mobile Safari/537.36"
}

class ExtCover:
    """
    Represents an external cover art.
    """
    artist: str
    title: str
    source: Source
    cover_url: str
    ext: str
    confidence: float = 1

    def __init__(self, artist: str, title: str, source: Source,
                 cover_url: str, ext: str, confidence: float = 1) -> None:
        self.artist = artist
        self.title = title
        self.source = source
        self.cover_url = cover_url
        self.ext = ext
        self.confidence = confidence
        self._buffer = None

    def get_buffer(self) -> io.BytesIO | None:
        """
        Fetches the cover art as a buffer and caches it
        """
        if self._buffer:
            return self._buffer

        r = requests.get(self.cover_url, headers=DEFAULT_HEADERS, timeout=60)

        if r.ok:
            self._buffer = io.BytesIO(r.content)
            return self._buffer
        
    def get_buffer_size(self):
        buffer = self.get_buffer()

        if buffer:
            return buffer.getbuffer().nbytes

    def download(self, path: str):
        """
        Download the cover art
        """
        buffer = self.get_buffer()

        if buffer:
            with open(path, 'wb') as f:
                f.write(buffer.getbuffer())

class Cover:
    """
    Represents a cover art file
    """
    path: str

    def __init__(self, path: str):
        self.path = path
        self.img = Image.open(self.path)

    @property
    def size(self):
        return os.path.getsize(self.path)

    @property
    def shape(self):
        return self.img.size

    def compare(self, ext_cover: ExtCover):
        """
        Compare similarity between an external cover art
        """
        hash_0 = imagehash.average_hash(self.img)
        buffer = ext_cover.get_buffer()

        if buffer:
            img = Image.open(buffer)
            hash_1 = imagehash.average_hash(img)

            return hash_0 - hash_1
    
    def delete(self):
        """
        Delete the current cover art file
        """
        os.remove(self.path)

    def backup(self):
        """
        Renames the current cover art to a .bk file
        """
        i = 0
        new_name = self.path + '.bk'
        while os.path.exists(new_name):
            i += 1
            name, ext = os.path.splitext(self.path)
            new_name = name + str(i) + ext + '.bk'
        os.rename(self.path, new_name)
        self.path = new_name



from dataclasses import dataclass
from coverdl.providers.source import Source

@dataclass
class Options:
    path: list[str]
    providers: list[Source]
    cover_name: str
    cache: str
    tags: list[str]
    recursive: bool
    upgrade: bool
    max_size: float
    max_upgrade_size: float
    strict: bool
    replace_non_square: bool
    max_hamming_distance: int
    silence_warnings: bool
    delete_old_covers: bool



from .base import Provider
from .deezer import DeezerProvider
from .itunes import ITunesProvider
from .apple_music import AppleMusicProvider
from .discogs import DiscogsProvider

providers: list[Provider] = [
    AppleMusicProvider(),
    DeezerProvider(),
    DiscogsProvider(),
    ITunesProvider()
]



import click

class Console:
    def __init__(self, silence_warnings: bool) -> None:
        self.silence_warnings = silence_warnings

    def warn(self, message: str):
        if self.silence_warnings:
            return
        click.echo(f"{click.style('Warn:', fg='yellow')} {message}")

    def error(self, message: str):
        click.echo(f"{click.style('Error:', fg='red')} {message}")

    def echo(self, *args, **kwargs):
        click.echo(*args, **kwargs)



from coverdl.console import Console
from coverdl.options import Options
from coverdl.providers import Provider
from coverdl.cache import Cache
from coverdl.cover import Cover, ExtCover
from coverdl.utils import get_cover, IMAGE_EXTENSIONS
from coverdl.metadata import get_metadata_from_path
from coverdl.exceptions import MetadataNotFound, MissingMetadata, TriesExceeded, ProviderRequestFailed
from requests.exceptions import Timeout
from mutagen import MutagenError
import click
import os


class UpgradeService:
    def __init__(self, path_locations: list[str], options: Options, providers: list[Provider]) -> None:
        self.console = Console(options.silence_warnings)
        self.cache = Cache(options.cache)
        self.path_locations = path_locations
        self.options = options
        self.providers = providers
        self.current_path = None

    def upgrade(self):
        for path in self.path_locations:
            self.current_path = path
            self._upgrade_cover_art(path)

        self.cache.save()

    def _get_results_from_providers(self, artist: str, album: str) -> list[ExtCover]:
        results: list[ExtCover] = []

        for provider in self.providers:
            try:
                results = results + provider.get_covers(artist, album)
            except ProviderRequestFailed as e:
                self.console.warn(f"Failed to fetch cover art data from provider: {e.args[0].value}. Got error: {e.args[1]}")
            except Timeout:
                self.console.warn(f"Provider {provider.SOURCE.value} timed out while fetching covers")

        return results

    def _meets_hamming_distance(self, cover: Cover, candidate: ExtCover) -> bool | None:
        if candidate.ext not in IMAGE_EXTENSIONS:
            self.console.warn(f"Image {candidate.ext} not allowed as valid image format.")
            return

        hamming_distance = None
        try:
            hamming_distance = cover.compare(candidate)
        except Timeout:
            self.console.error(f"Timed out while downloading cover art for {click.style(self.current_path, bold=True)}.")
            return

        if hamming_distance is None:
            self.console.warn(f"Failed to calculate hamming distance for candidate {candidate.cover_url}. Skipping.")
            return

        similarity_check = hamming_distance == 0 if self.options.strict else hamming_distance <= self.options.max_hamming_distance

        if not similarity_check:
            self.console.warn(f"Cover candidate ({candidate.cover_url}) for {click.style(self.current_path, bold=True)} " \
                        f"does not meet similarity requirements (hamming distance = {hamming_distance}, needs <= {self.options.max_hamming_distance}). Skipping.")
            
        return similarity_check

    def _find_best_candidate(self, cover: Cover, results: list[ExtCover]) -> ExtCover | None:
        cover_size = cover.size / 1000000

        for cover_candidate in results:
            candidate_buffer_size = cover_candidate.get_buffer_size() / 1000000

            if candidate_buffer_size > self.options.max_upgrade_size:
                self.console.warn(f"Cover candidate ({cover_candidate.cover_url}) for {click.style(self.current_path, bold=True)} " \
                    f"exceeds max_upgrade_size ({round(candidate_buffer_size)}M / {self.options.max_upgrade_size}M). Skipping.")
                continue

            if candidate_buffer_size <= cover_size:
                self.console.warn(f"Cover candidate ({cover_candidate.cover_url}) for {click.style(self.current_path, bold=True)} " \
                    f"is less than or equivalent in size ({candidate_buffer_size}M / {cover_size}M). Skipping.")
                continue

            if self._meets_hamming_distance(cover, cover_candidate):
                return cover_candidate

    def _delete_or_backup(self, cover: Cover):
        if self.options.delete_old_covers:
            cover.delete()
            self.console.echo(f"Deleted {cover.path}")
        else:
            old_cover = cover.path
            cover.backup()
            self.console.echo(f"Renamed {old_cover} to {cover.path}")

    def _upgrade_cover_art(self, path: str):
        cover = get_cover(path)

        if self.cache.has(os.path.abspath(path)):
            self.console.warn(f"{click.style(path, bold=True)} exists in cache. It'll be skipped.")
            return

        if not cover:
            self.console.warn(f"{click.style(path, bold=True)} does not have cover. Skipping.")
            return

        cover_size = cover.size / 1000000

        if cover_size > self.options.max_size:
            self.console.warn(f"{click.style(cover, bold=True)} exceeds --max-size ({round(cover_size, 2)}M / {self.options.max_size}M). Skipping.")
            return

        try:
            self.console.echo(f"Fetching metadata from {click.style(path, bold=True)}")
            metadata = get_metadata_from_path(path)
        except (MutagenError, TriesExceeded, MissingMetadata, MetadataNotFound):
            self.console.error(f"Failed to fetch metadata from {path}")
            self.cache.add(os.path.abspath(path))
            return

        results: list[ExtCover] = self._get_results_from_providers(metadata.artist, metadata.album)

        if len(results) == 0:
            self.console.error(f"No suitable upgradable cover art could be found for {click.style(path, bold=True)} (all providers returned no results)")
            self.cache.add(os.path.abspath(path))
            return
        
        candidate: ExtCover | None = None

        if self.options.replace_non_square and cover.shape[0] != cover.shape[1]:
            self.console.echo(f"{cover.path} is non-square. Replacing...")
            candidate = results[0]
        else:
            candidate: ExtCover | None = self._find_best_candidate(cover, results)

        if not candidate:
            self.console.error(f"No suitable cover art could be found for {click.style(path, bold=True)} (exhausted all candidates)")
            self.cache.add(os.path.abspath(path))
            return
        
        self._delete_or_backup(cover)

        target = os.path.join(path, self.options.cover_name + candidate.ext)
        candidate.download(target)

        click.echo(f"{click.style('Successfully', fg='green')} saved new cover art: {target}")
        self.cache.add(os.path.abspath(path))



import sys
import click
from coverdl import __version__
from coverdl.console import Console
from coverdl.providers import providers
from coverdl.providers.source import Source
from coverdl.upgrade import UpgradeService
from coverdl.download import DownloadService
from coverdl.utils import get_album_paths
from coverdl.options import Options

@click.command(context_settings={'show_default': True})
@click.option('-p', '--provider',
              type=Source, multiple=True, default=[Source.APPLE_MUSIC, Source.DEEZER],
              help='The provider to download cover art from. Multiple providers can be specified by passing this option more than once (by order of preference).')
@click.option('--cover-name',
              default='cover',
              help='The filename for the cover art image.')
@click.option('--cache',
              help='Specify a cache file to prevent repeat upgrades')
@click.option('-t', '--tag',
              multiple=True,
              help='A tag is an attribute unique to your release. e.g: remastered, expanded. Multiple tags can be specified.')
@click.option('-r', '--recursive',
              is_flag=True,
              help='Enables recursive download. This will traverse the path you give it and download cover art for albums that don\'t have them.')
@click.option('-u', '--upgrade',
              is_flag=True,
              help='Upgrades existing cover art to better quality equivalents.')
@click.option('--max-size',
              type=float, default=10,
              help='Already existing cover art that exceeds the max size will not be considered for upgrade (unit must be in MBs).')
@click.option('--max-upgrade-size',
              type=float, default=15,
              help='Upgrade candidates exceeding this size will not be considered (unit must be in MBs).')
@click.option('--strict',
              help='Enables strict mode (for upgrades). Ensures that only near-perfect comparisons will be upgraded.')
@click.option('--replace-non-square',
              is_flag=True,
              help='Replace non-square cover art. Must be used alongside --upgrade')
@click.option('--max-hamming-distance',
            type=int, default=8,
            help='Specifies the maximum hamming distance used for upgrades. Setting this to 0 is the equivalent of using --strict')
@click.option('--silence-warnings',
              is_flag=True)
@click.option('--delete-old-covers',
              is_flag=True)
@click.argument('path', type=click.Path(exists=True), nargs=-1)
@click.version_option(__version__)
def coverdl(path: list[str],
            provider: list[Source],
            cover_name: str,
            cache: str,
            recursive: bool,
            upgrade: bool,
            tag: list[str],
            max_size: float,
            max_upgrade_size: float,
            strict: bool,
            replace_non_square: bool,
            max_hamming_distance: int,
            silence_warnings: bool,
            delete_old_covers: bool):
    options = Options(
        path=path,
        providers=provider,
        cover_name=cover_name,
        cache=cache,
        tags=tag,
        recursive=recursive,
        upgrade=upgrade,
        max_size=max_size,
        max_upgrade_size=max_upgrade_size,
        strict=strict,
        replace_non_square=replace_non_square,
        max_hamming_distance=max_hamming_distance,
        silence_warnings=silence_warnings,
        delete_old_covers=delete_old_covers
    )

    console = Console(options.silence_warnings)
    
    # Get and sort the providers based on user preference
    selected_providers = list(filter(lambda p: p.SOURCE in provider, providers))
    selected_providers.sort(key=lambda p: options.providers.index(p.SOURCE))

    if options.recursive and options.tags:
        console.error("--recursive and --tag cannot be used together.")
        return

    path_locations: list[str] = []
    if not sys.stdin.isatty():
        stdin_text = click.get_text_stream('stdin')
        path_locations = list(stdin_text.read().rstrip().split('\n'))
    else:
        if options.recursive and len(options.path) != 1:
            console.error("Please specify (one) path for recursive search.")
            return
        path_locations = get_album_paths(options.path[0], must_have_cover=options.upgrade) if options.recursive else list(options.path)

    if options.upgrade:
        upgrade_service = UpgradeService(path_locations, options, selected_providers)
        upgrade_service.upgrade()
    else:
        download_service = DownloadService(path_locations, options, selected_providers)
        download_service.download()



import os
import click
from requests import HTTPError, Timeout
from mutagen import MutagenError
from coverdl.console import Console
from coverdl.cover import ExtCover
from coverdl.exceptions import MetadataNotFound, ProviderRequestFailed, TriesExceeded
from coverdl.metadata import get_metadata_from_path
from coverdl.options import Options
from coverdl.providers.base import Provider
from coverdl.utils import IMAGE_EXTENSIONS, has_cover


class DownloadService:
    def __init__(self, path_locations: list[str], options: Options, providers: list[Provider]) -> None:
        self.console = Console(options.silence_warnings)
        self.path_locations = path_locations
        self.options = options
        self.providers = providers

        self.completed = 0
        self.failed = 0

    def download(self):
        for path in self.path_locations:
            self._download_cover_art(path)

        self.console.echo()
        self.console.echo(f"{click.style('Completed:', bold=True)} {self.completed}, {click.style('Failed:', bold=True)} {self.failed}")

    def _get_results_from_providers(self, artist: str, album: str) -> list[ExtCover]:
        results: list[ExtCover] = []

        for provider in self.providers:
            try:
                results = results + provider.get_covers(artist, album)

                if len(results) != 0:
                    # There is no need to fetch from the next provider if we already have results.
                    break
            except ProviderRequestFailed as e:
                self.console.warn(f"Failed to fetch cover art data from provider: {e.args[0].value}. Got error: {e.args[1]}")
            except Timeout:
                self.console.warn(f"Provider {provider.SOURCE.value} timed out while fetching covers")

        return results

    def _download_cover_art(self, path: str):
        self.console.echo(f"Fetching metadata from {path}")

        try:
            metadata = get_metadata_from_path(path)
        except (TriesExceeded, MetadataNotFound):
            self.console.error(f"Failed to fetch sufficient metadata from album: {click.style(path, bold=True)}")
            self.failed += 1
            return
        except MutagenError:
            self.console.error(f"Failed to read file: {click.style(path, bold=True)}, possibly invalid or corrupt")
            self.failed += 1
            return
        
        if has_cover(path):
            self.console.warn(f"{click.style(path, bold=True)} already has a cover. Skipping.")
            return
        
        if not metadata:
            self.console.error(f"Could not fetch metadata from path: {click.style(path, bold=True)}")
            self.failed += 1
            return
        
        results = self._get_results_from_providers(metadata.artist, metadata.album)

        if len(results) == 0:
            self.console.error(f"No suitable cover art could be found for {click.style(path, bold=True)}")
            self.failed += 1
            return
        
        cover: ExtCover | None = None

        for result in results:
            if result.ext not in IMAGE_EXTENSIONS:
                self.console.warn(f"Cover image format not allowed {result.ext}. Skipping.")
                continue
            cover = result

        if not cover:
            self.console.error(f"Failed to find suitable cover art for {click.style(path, bold=True)}.")
            self.failed += 1
            return
        
        try:
            dir_path = path if os.path.isdir(path) else os.path.dirname(path)
            cover.download(os.path.join(dir_path, self.options.cover_name + cover.ext))
            click.echo(f"{click.style('Successfully', fg='green')} downloaded cover art for {click.style(path, bold=True)}")
            self.completed += 1
        except HTTPError as e:
            self.console.error(f"Failed to download cover art for {click.style(path, bold=True)}, got error: {e.response.status_code}")
        except Timeout:
            self.console.error(f"Timed out while downloading cover art for {click.style(path, bold=True)}")

